import Graph;
import java.util.Stack;
import java.util.Vector;
import java.util.Random;

/*
 *
 * KombGraph
 *
 */

public class KombGraph extends Graph {
	static boolean verbose = false;	// Ausgabe von Ablauf-information
	static Vector Tasks = new Vector();		// Die Liste aller Tasks
	static Vector ENV = new Vector();		// ENV enthlt alle definierten Ausdrcke
	static Random RND = new Random();
	public char type;	  //S : String/Symbol, I = int, R = real, K = Kombinator , A = Application
	public char Emark;	  //E = Eval, R = Ready, B = Busy 
	public int waitsFor;
	public Stack WQ;

	public KombGraph(String s) {
		this.datum = s;
		this.type = 'S';
		this.Emark = 'R';
		this.WQ = new Stack();
		this.waitsFor=0;
	}


	public KombGraph(Kombinator k) {
		this.datum = k;
		this.type = 'K';
		this.Emark = 'R';
		this.WQ = new Stack();
		this.waitsFor=0;
	}

	public KombGraph(KombGraph l, KombGraph r) {
		this.addLeft(l);
		this.addRight(r);
		this.datum = null;
		this.type = 'A';
		this.Emark = 'E';
		this.WQ = new Stack();
		this.waitsFor=0;
	}
	  
	public KombGraph(Integer i) {
		this.datum = i;
		this.type = 'I';
		this.Emark = 'R';
		this.WQ = new Stack();
		this.waitsFor=0;
	}

	public KombGraph(Float r ) {
		this.datum = r;
		this.type = 'R';
		this.Emark = 'R';
		this.WQ = new Stack();
		this.waitsFor=0;
	}

	public String toString() {
		return ((Graph) this).toString(0);
	}

	public Kombinator spine() throws Exception{
		if (this.type == 'K')
			return (Kombinator) this.datum;
		else if (this.type == 'A')
			return ((KombGraph) this.left()).spine();
		else throw new Exception("No valid Combinator: " + this.toString()); // Kein Kombinator als Spine gefunden !!
	}

	public Stack LeftAncestorStack(Stack s) {
		if (this.type != 'A')  //wenn der Knoten keine @ mehr ist, ist der stack fertig
			return s;
		else {s.push(this);
			  return ((KombGraph) this.left()).LeftAncestorStack(s);
		}		
	}

	//set berschreibt einen Knoten mit einem anderen bestehenden
	public void set(KombGraph g){
		this.datum = g.datum;
		this.type = g.type;
		this.Emark = g.Emark;
		//this.WQ = g.WQ;
		//this.waitsFor = g.waitsFor;
		this.addLeft(g.left());
		this.addRight(g.right());
	}

	//red fhrt die eigentliche Graphreduktion nach den Regeln der Kombinatorlogik durch
	public void red() throws Exception{
	  if (this.type == 'A'){
		
			Kombinator k = this.spine();   //liegt ein wohlgeformter Graph vor?
			if (verbose) System.out.println("spine: " + k.toString());

			Stack stack = this.LeftAncestorStack(new Stack());
			KombGraph top = (KombGraph) stack.pop();
			if (verbose) 
			  {System.out.println("top: " + top.toString());
			   System.out.println(k.name);
			  }

			if (k.name.equals("I"))	  //Reduktion Kombinator I
			 {	
				KombGraph tmp = (KombGraph) top.right();
				top.set( tmp);
			    if (verbose) System.out.println("red I: " + this.toString());
			}
			else if (k.name.equals("K")) //Reduktion Kombinator K
			 {	
				KombGraph tmp = (KombGraph) top.right();
				((KombGraph) stack.pop()).set(tmp);
				if (verbose) System.out.println("red K: " + this.toString());
			}
			else if (k.name.equals("S")) //Reduktion Kombinator S
			 {	KombGraph x = (KombGraph) top.right();
				KombGraph y = (KombGraph) ((KombGraph) stack.pop()).right();
				KombGraph modified = (KombGraph) stack.pop();
				KombGraph z = (KombGraph) modified.right();
				KombGraph vorne = new KombGraph(x,z);
				KombGraph hinten = new KombGraph(y,z);
				modified.set(new KombGraph(vorne,hinten));
			    if (verbose) System.out.println("red S: " + this.toString());}
			else if (k.name.equals("ERROR")) //Reduktion ERROR Message
			 {	KombGraph Message = (KombGraph) top.right();
				System.out.println("ERROR: " + Message.toString());
				top.Emark = 'R';
				top.set(null);
				throw new Exception("ERROR: " + Message.toString());
				//Noch keine weitere Fehlerbehandlung implementiert !!
			 }
			else if	(k.name.equals("+")) //Reduktion + Integer
			 {	
				KombGraph x = (KombGraph) top.right();
				KombGraph result = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) result.right();
				if ((x.type == 'I') && (y.type == 'I'))
				{
					Integer value =  new Integer(((Integer) x.datum).intValue() + ((Integer) y.datum).intValue());
					result.set(new KombGraph(value));
				}
				else {
					x.subEval(result);
					y.subEval(result); 
					//result.remTask();
				}


				if (verbose) System.out.println("red +: " + this.toString());
				
			 }
			else if	(k.name.equals("-")) //Reduktion - Integer
			 {	
				KombGraph x = (KombGraph) top.right();
				KombGraph result = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) result.right();
				if ((x.type == 'I') && (y.type == 'I'))
				{
					Integer value =  new Integer(((Integer) x.datum).intValue() - ((Integer) y.datum).intValue());
					result.set(new KombGraph(value));
				}
				else {
					x.subEval(result);
					y.subEval(result);
					//result.remTask();
				}


				if (verbose) System.out.println("red -: " + this.toString());
				
			 }


			else if	(k.name.equals("*")) //Reduktion * Integer
			 {	
				KombGraph x = (KombGraph) top.right();
				KombGraph result = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) result.right();
				if ((x.type == 'I') && (y.type == 'I'))
				{
					Integer value =  new Integer(((Integer) x.datum).intValue() * ((Integer) y.datum).intValue());
					result.set(new KombGraph(value));
				}
				else {
					x.subEval(result);
					y.subEval(result);
					//result.remTask();
				}


				if (verbose) System.out.println("red *: " + this.toString());
				
			 }

		     else if	(k.name.equals("/")) //Reduktion / Integer
			 {	
				KombGraph x = (KombGraph) top.right();
				KombGraph result = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) result.right();
				if ((x.type == 'I') && (y.type == 'I'))
				{
					Integer value =  new Integer(((Integer) x.datum).intValue() / ((Integer) y.datum).intValue());
					result.set(new KombGraph(value));
				}
				else {
					x.subEval(result);
					y.subEval(result);
				
				}


				if (verbose) System.out.println("red /: " + this.toString());
				
			 }


			else if	(k.name.equals("=")) //Reduktion = Integer
			 {	
				KombGraph x = (KombGraph) top.right();
				KombGraph result = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) result.right();
				if ((x.type == 'I') && (y.type == 'I'))
				{
					if ( ((Integer) x.datum).intValue() == ((Integer) y.datum).intValue() )
						Integer value =  new Integer(1);
					else  value =  new Integer(0);
					result.set(new KombGraph(value));
				}
				else {
					x.subEval(result);
					y.subEval(result);
					//result.remTask();
				}


				if (verbose) System.out.println("red =: " + this.toString());
				
			 }

			else if (k.name.equals("car"))	  //Reduktion car
			 {	KombGraph x = (KombGraph) top.right();
				if (x.Emark == 'R'){
				  try {
					Stack l = x.LeftAncestorStack(new Stack());
					KombGraph tp = (KombGraph) l.pop();

					KombGraph tmp = (KombGraph) tp.left();
					top.set(tmp);
					top.Emark = 'R';
			      }
			      catch (Exception e){
					throw new Exception("Error: car: " + ((KombGraph) top.right()).toString());
				  }
			      if (verbose) System.out.println("red car: " + this.toString());
				}
				else x.subEval(top);
			}


			else if (k.name.equals("cdr"))	  //Reduktion car
			 {	KombGraph x = (KombGraph) top.right();
				if (x.Emark == 'R'){
				
				try {
					Stack l = ((KombGraph) top.right()).LeftAncestorStack(new Stack());
					KombGraph tp = (KombGraph) l.pop();
					tp.set((KombGraph) tp.right());

					KombGraph tmp = (KombGraph) top.right();
					top.set(tmp);
					top.Emark = 'R';
				}
				catch (Exception e){
					throw new Exception("Error: cdr: " + ((KombGraph) top.right()).toString());
				}
				}
				else x.subEval(top);
			    if (verbose) System.out.println("red cdr: " + this.toString());
			}

			else if (k.name.equals("quote")) {	  //Reduktion quote
				top.set((KombGraph) top.right());
				top.Emark = 'R';

				if (verbose) System.out.println("red quote: " + this.toString());
			}

			else if (k.name.equals("eval"))	  //Reduktion eval
			 {	KombGraph x = (KombGraph) top.right();
				if (x.Emark == 'R'){
					top.set(x);
					top.Emark = 'E';
					if (verbose) System.out.println("red eval: " + this.toString());
				}
				else x.subEval(top);

				
				
			}
		  


			else if (k.name.equals("atomp"))	  //Reduktion atomp
			 {	KombGraph x = (KombGraph) top.right();
				if (x.Emark == 'R'){
					Integer ret = new Integer(0);
					if (x.type != 'A'  )
						ret = new Integer(1);
					top.set(new KombGraph(ret));				
					if (verbose) System.out.println("red atomp: " + this.toString());
				}
				else x.subEval(top);
			}



			else if	(k.name.equals("eq")) //Reduktion eq KombGraphs
			 {	
				KombGraph x = (KombGraph) top.right();
				KombGraph result = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) result.right();
				Integer value = new Integer(0);
				if (x.equal(y))	{
					System.out.println("1");
					value =  new Integer(1);
				}
				
				result.set(new KombGraph(value));

				if (verbose) System.out.println("red eq: " + this.toString());
				
			 }


			else if (k.name.equals("not")) //Reduktion Kombinator not
			 	{	
				KombGraph x = (KombGraph) top.right();
				//KombGraph result = (KombGraph) stack.pop();
				//KombGraph y = (KombGraph) result.right();
				if (x.type == 'I')
				{	int res = 0;
					if ( ((Integer) x.datum).intValue() == 0)
						res = 1;

					Integer value =  new Integer(res);
					top.set(new KombGraph(value));
				}
				else {
					x.subEval(top);
					
				}
			}



			else if	(k.name.equals("or")) //Reduktion and (0 v 1)
			 {	int test=1;
				KombGraph x = (KombGraph) top.right();
				KombGraph result = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) result.right();
				if ((x.type == 'I') && (y.type == 'I'))
				{	
					if (((Integer) x.datum).intValue() == 1 || ((Integer) y.datum).intValue() == 1) test = 1;
					result.set(new KombGraph(new Integer(test)));
				}
				else if (x.type == 'I')
				{	
					if (((Integer) x.datum).intValue() == 1) test = 1;
					result.set(new KombGraph(new Integer(test)));

					y.remTask();

				}
				else if (y.type == 'I')
				{	
					if (((Integer) y.datum).intValue() == 1) test = 1;
					result.set(new KombGraph(new Integer(test)));
					x.remTask();
				}


				else {
					if (this.waitsFor == 0){
						x.newTask();
						y.newTask();
						this.Emark = 'E';
						this.waitsFor = this.waitsFor + 2;
						
					}
					else return;
					
				}


				if (verbose) System.out.println("red or: " + this.toString());
				
			 }



			else if	(k.name.equals("and")) //Reduktion and (0 v 1)
			 {	int test=0;
				KombGraph x = (KombGraph) top.right();
				KombGraph result = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) result.right();
				if ((x.type == 'I') && (y.type == 'I'))
				{	
					if (((Integer) x.datum).intValue() == 1 && ((Integer) y.datum).intValue() == 1) test = 1;
					result.set(new KombGraph(new Integer(test)));
				}
				else if (x.type == 'I')
				{	
					if (((Integer) x.datum).intValue() == 0) test = 0;
					result.set(new KombGraph(new Integer(test)));

					y.remTask();

				}
				else if (y.type == 'I')
				{	
					if (((Integer) y.datum).intValue() == 0) test = 0;
					result.set(new KombGraph(new Integer(test)));
					x.remTask();
				}


				else {
					if (this.waitsFor == 0){
						x.newTask();
						y.newTask();
						this.Emark = 'E';
						this.waitsFor = this.waitsFor + 2;
						
					}
					else return;
					
				}


				if (verbose) System.out.println("red and: " + this.toString());
				
			 }


			else if	(k.name.equals("if")) //Reduktion if Integer
			 {	KombGraph test = (KombGraph) top.right();
				KombGraph thn = (KombGraph) ((KombGraph) stack.pop()).right();
				KombGraph modified = (KombGraph) stack.pop();
				KombGraph els = (KombGraph) modified.right();

				if ((test.type == 'I'))
				{
					if ( ((Integer) test.datum).intValue() == 1 )
					  modified.set(thn);
					else  modified.set(els);
				}
				else {
					test.subEval(modified);
					
				}


				if (verbose) System.out.println("red =: " + this.toString());
				
			 }


		   	else if	(k.name.equals("C")) //Reduktion C Kombinator
			 {	
				KombGraph x = (KombGraph) top.right();
				KombGraph y = (KombGraph) ((KombGraph) stack.pop()).right();
				KombGraph modified = (KombGraph) stack.pop();
				KombGraph z = (KombGraph) modified.right();
				KombGraph vorne = new KombGraph(x,z);
				modified.set(new KombGraph(vorne,y));

			    if (verbose) System.out.println("red C: " + this.toString());
			 } 

			else if	(k.name.equals("Y")) //Reduktion Y Kombinator Y f := f (Y f) 
			 {	
				KombGraph f = (KombGraph) top.right();
				top.set(new KombGraph(f,new KombGraph(new KombGraph(new Kombinator("Y")),f)));
			
			    if (verbose) System.out.println("red Y: " + this.toString());
			 }

			else if	(k.name.equals("B")) //Reduktion B Kombinator
			 {	
				KombGraph x = (KombGraph) top.right();
				KombGraph y = (KombGraph) ((KombGraph) stack.pop()).right();
				KombGraph modified = (KombGraph) stack.pop();
				KombGraph z = (KombGraph) modified.right();
				KombGraph hinten = new KombGraph(y,z);
				modified.set(new KombGraph(x,hinten));
			    if (verbose) System.out.println("red B: " + this.toString());
			 }
			else if (k.name.equals("cons")) //Reduktion Cons
			 {	KombGraph x = (KombGraph) top.right();
				KombGraph root = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) root.right();
				
				if (x.Emark == 'R' && y.Emark == 'R'){
					if (y.type == 'A') {
					
						Stack s = y.LeftAncestorStack(new Stack());
						

						KombGraph tp = (KombGraph) s.pop();

						
						KombGraph tpleftclone = new KombGraph(null,null);
						tpleftclone.set((KombGraph) tp.left());

						

						KombGraph tmp = new KombGraph(x,tpleftclone);
						
						
						((KombGraph) tp.left()).set(tmp);
						
						
						root.set(y);
					}
					else {
						KombGraph tmp = new KombGraph(x,y);
						root.set(tmp);
					}

					
					root.Emark = 'R';
					return;
				}
				else {
					x.subEval(root);
					y.subEval(root);
					return;
				}
			}

			//Definition statischer Variablen

			else if (k.name.equals("define")) {
				String var = (String) ((Kombinator) top.right().datum).name;
				KombGraph root = (KombGraph) stack.pop();
				KombGraph val = (KombGraph) root.right();
				define(var,val);
				root.Emark = 'R';
			}

			/*// Variablen Lookup
			else if (k.name.equals("DEF")) {
				if (verbose) System.out.println("def rein");

				String key = (String) top.right().datum;
				top.set(lookUp(key));
				top.Emark = 'R';
				if (verbose) System.out.println("red DEF: " + this.toString());


			}
			*/

			else if	(k.name.equals("P")) //Reduktion Proemial-Kombinator P
				// Reduktion nach Schema: P (f xf) (g xg) ==> P parEval(f xf) parEval(g xg)
			 {	
				KombGraph x = (KombGraph) top.right();
				KombGraph root = (KombGraph) stack.pop();
				KombGraph y = (KombGraph) root.right();
				x.parEval(root);
				y.parEval(root);
				//k.name = "quote";
				root.Emark = 'R';
			    if (verbose) System.out.println("red P: " + this.toString());
			 }
			


			// Wenn alles andere fehlschlgt:
			else 
			{

				if (verbose) System.out.println("User defined Kombinator " + k.name);

				
				((KombGraph) top.left()).set(lookUp(k.name));
				((KombGraph) top.left()).Emark = 'R';
				if (verbose) System.out.println("red Lookup: " + this.toString());



			}
			

		
		
	  }
	  else	 {  if (this.type == 'K'){	   // Int, real, string are in normal form

				if (verbose) System.out.println("User defined Variable " + ((Kombinator) this.datum).name);

				
				this.set(lookUp(((Kombinator) this.datum).name));
				this.Emark = 'R';
				if (verbose) System.out.println("red Lookup: " + this.toString());
			   }

	  }
	}




	//makeWait setzt einen Knoten in den Wartezustand
	public void makeWait(){
		this.waitsFor = this.waitsFor +1;
		this.Emark = 'B';
		this.remTask();
	}

	// makeUnwait setzt den waitsFor-Zhler um 1 herunter, 
	// wenn der Knoten auf niemand mehr wartet, wird weiter evaluiert
	public void makeUnwait(){
		//System.out.println(this.waitsFor + "");
		this.waitsFor = this.waitsFor - 1;
		//System.out.println(this.waitsFor + "");

		
		if (this.waitsFor == 0)
		 {this.Emark ='E';
		  this.newTask();
		}
	}

	// wakeUp weckt alle wartenden Knoten wieder auf,
	// wenn ein Knoten fertig evaluiert ist
	public void wakeUp()  {
		while (! this.WQ.empty())
		{((KombGraph) this.WQ.pop()).makeUnwait();
		}
	}

	


	//SubEval startet eine synchronisierte Subevaluation
	public void subEval(KombGraph root){
		 if (this.Emark == 'R') 
		   {return;}
		 else if (this.Emark == 'B')
		   { root.makeWait();
			 this.WQ.push(root); 
			 
		   }
		 else 
		   { root.makeWait();
			 this.WQ.push(root);
			 //this.Emark = 'B';
			 
			 this.newTask();

		   } 
	}

	
	// parEval startet eine unsynchronisierte Subevaluation
	// zb. fr den P-Kombinator
	public void parEval(KombGraph root){
		 if (this.Emark == 'R') 
		   {return;}
		 /*else if (this.Emark == 'B')
		   { //root.makeWait();
			 this.WQ.push(root); 
		   } */
		 else 
		   { 
			 //this.Emark = 'B';
			 //this.WQ.push(root);
			 this.newTask();

		   } 
	}


	//newTask lst fr den Konten eine separate Evaluation aus
	public void newTask(){
		Tasks.addElement(this);
	}

	//remTask nimmt einen Knoten aus der Taskliste
	public void remTask(){
		if (Tasks.removeElement(this)) return;
		else 
		{	if (verbose) System.out.println("Knoten nicht in der Taskliste");
			return;
		
		}
	}

	static int MaxTasks; 
	static int StepCount;

	// Scheduler realisert das multithreaded Abarbeiten der Tasks
	static void Scheduler() throws Exception{
		MaxTasks = 0; StepCount = 0;
		while (! (Tasks.size() == 0)) 
		{	StepCount++;
			if (verbose) System.out.println(Tasks.toString());
			if (Tasks.size() > MaxTasks) MaxTasks = Tasks.size();
			Double ind = new Double(RND.nextDouble()*Tasks.size());
			int index = ind.intValue();
			((KombGraph) Tasks.elementAt( index )).evalstep();
			
			if (verbose) System.out.println(Tasks.toString());
		}
	}


	//evalstep fhrt einen Reduktionsschritt durch
	public void evalstep() throws Exception{
		/*if (!(this.type == 'A')){
			this.wakeUp();
			this.remTask();	 }
		else */if (this.Emark == 'B')	 {}
			
		else {
			this.red();
			if (this.Emark == 'R') 
			{
				this.wakeUp();
				this.remTask();
			} 
			else return;
		} 

	}


	//eval liefert die Normalform eines Knoten
	public KombGraph eval() throws Exception{
	   	//Tasks = null;
		//System.out.println(this.toString());
		this.newTask();
		Scheduler();
		return this;
	   
	}
	
	// Definition von Variablen, die statisch gltig bleiben!
	// Bindung ist dynamisch !
	// berschreiben alter Werte ist mglich

	public static void define(String var, KombGraph val) {
		//boolean found = false;
		try{
			lookUp(var)	;
			for (int n=0; (n < ENV.size() ); n++) {
				if (((envEntry) ENV.elementAt(n)).key.equals(var)){
					ENV.removeElementAt(n);
					ENV.addElement(new envEntry(var,val));
				}
			}
		}
		catch (Exception e){
			ENV.addElement(new envEntry(var,val));
		}
	}

	public static KombGraph lookUp(String var) throws Exception{
		//System.out.println("lookup rein: " + var);
		for (int n=0; (n < ENV.size() ); n++)
			{if (((envEntry) ENV.elementAt(n)).key.equals(var))
				return ((envEntry) ENV.elementAt(n)).value;
		}

		KombGraph res = new KombGraph(new KombGraph(new Kombinator("ERROR")),new KombGraph("Variable " + var + " not found"));
		res.Emark = 'R';
		throw new Exception("Variable " + var + " not defined");
		
	}

	//equal compares 2 KombGraphs
	public boolean equal(KombGraph that) {
		//System.out.println(this.toString());
		//System.out.println(that.toString());
		//System.out.println(this.type);
		//System.out.println(that.type);
		//System.out.println(this.datum.toString());
		//System.out.println(that.datum.toString());
		if (this.type == that.type){
			if (this.type == 'I' ){ 
				if 
					(((Integer) this.datum).intValue() == ((Integer) that.datum).intValue())
					return true;
				else return false;}
			else if (this.type == 'S' ){
				 if 
					(((String) this.datum).equals((String) that.datum))
					return true;
				 else return false;

			}
			else if (this.type == 'K') {
				if ( ((Kombinator) this.datum).name.equals(((Kombinator) that.datum).name) )
					return true;
				else return false;

			}
			else if (this.type == 'R'){
				System.out.println("Reals not yet implemented");
				return false;
			}

			// nur bei Applikationen type == 'A'
			else if ( ((KombGraph) this.right()).equal((KombGraph) that.right()) ){ 
					System.out.println("R = R");
					if  
					  ( ((KombGraph) this.left()).equal((KombGraph) that.left())  ) {
						System.out.println("L = L");
						return true;
					}
					else return false;
			}

			else return false;
		}
		else return false;
	}

	public String pretty() {
		String s = this.toString();
		String tmp = "";
		for (int i=0 ; i < s.length(); i++){
			if (s.substring(i).startsWith("(DEF")){

			}	
			else {
				tmp = tmp + s.charAt(i);
			}
		}
		return tmp;

	}


}

class envEntry {
	String key;
	KombGraph value;

    public envEntry(String k,KombGraph v){
		 key = k;
		 value = v;

	}

	public String toString(){
		return key + " := " + value.toString();
	}
}